
#ifdef PROM
BASE = 0
#define BANNER "@(#)RD58 P8E/P8N 0.2 PROM 970219"
#else
BASE = 0xD000
#define BANNER "@(#)RD58 P8E/P8N 0.2 RAM 970219"
#endif

!----------------------------------------------------------------------
!
!	Convenience macros
!
#define BUF(n) : .rs n
#define BYTE   : .rs 1
#define WORD   : .rs 2

#define MKWORD(lo,hi) ((lo) | ((hi) << 8))

#define MASK(bit) (1 << (bit))

#define srl4 srl a ; srl a ; srl a ; srl a

#define imm_ahl(I) ld hl, (I) % 65536 ; ld a, (I) / 65536

#define load_ahl(A) ld hl, [A] ; ld a, [(A) + 2]
#define save_ahl(A) ld [A], hl ; ld [(A) + 2], a
#define load_abc(A) ld bc, [A] ; ld a, [(A) + 2]

!----------------------------------------------------------------------
!
!	I/O map
!
#define PIO   0x00
#define SIO   0x10
#define TMR   0x20
#define DA0   0x30
#define DA1   0x40
#define AD    0x50
#define OUT0  0x60
#define OUT1  0x70
#define OUT2  0x80
#define WD    0x90
#define MDM   0xA0

#define CSMEM 0xB0	/* P8E SMEM flipflop  */

#define ADATA 0
#define BDATA 1
#define ACTRL 2
#define BCTRL 3

#define TMRCTRL 3

#define AD_RSSI  (AD + 0)

#define DA_RFC   DA0
#define DA_TXPWR DA1

!----------------------------------------------------------------------
!
!	FX429
!
#define MDMDATA	2
#define MDMCTRL 3

#define MDM_TXENB	0x01
#define MDM_TXPAR	0x02
#define MDM_RXENB	0x04
#define MDM_RXFMT	0x08
#define MDM_TIMER	0xF0

#define MDM_XXX		0x00	/* Nothing enabled */

#define MDM_RXRDY	0x01
#define MDM_RXTRUE	0x02
#define MDM_DCD		0x04
#define MDM_TXRDY	0x08
#define MDM_TXIDL	0x10
#define MDM_TMRINT	0x20
#define MDM_SYNC	0x40
#define MDM_SYNT	0x80

!----------------------------------------------------------------------
!
!	8254 timer
!

#define TMR_0		0x00
#define TMR_1		0x40
#define TMR_2		0x80
#define TMR_INTTC	0x00	/* mode 0, intr on terminal cnt */
#define TMR_ONESHOT	0x02	/* mode 1, one shot */
#define TMR_RATEGEN	0x04	/* mode 2, rate generator */
#define TMR_SQWAVE	0x06	/* mode 3, square wave */
#define TMR_SWTRIGG	0x08	/* mode 4, s/w triggered strobe */
#define TMR_LSB		0x10	/* r/w counter LSB */
#define TMR_MSB		0x20	/* r/w counter MSB */
#define TMR_BOTH	0x30	/* r/w counter 16 bits, LSB first */

!----------------------------------------------------------------------
!
!	SIO defs
!

#define WR0_SEND_ABORT			0x08
#define WR0_RESET_ESCINT		0x10
#define WR0_CHANNEL_RESET		0x18
#define WR0_REARM_1STC			0x20
#define WR0_RESET_TXINT			0x28
#define WR0_ERROR_RESET			0x30
#define WR0_RETINT				0x38
#define WR0_RESET_RXCRC			0x40
#define WR0_RESET_TXCRC			0x80
#define WR0_RESET_UNDERRUN		0xc0

#define WR1_ESCINT_ENB			0x01
#define WR1_TXINT_ENB			0x02
#define WR1_STATUS_VECTOR		0x04
#define WR1_RXINT_DISABLE		0x00
#define WR1_INT_1STC			0x08
#define WR1_INT_ALLC			0x10
#define WR1_INT_ALLC_PNS		0x18
#define WR1_WAITREADY_RT		0x20
#define WR1_WAITREADY_FN		0x40
#define WR1_WAITREADY_ENB		0x80

#define WR3_RX_ENB				0x01
#define WR3_SYNCLOAD_INHIBIT	0x02
#define WR3_ADDRESS_SEARCH_MODE	0x04
#define WR3_RXCRC_ENB			0x08
#define WR3_ENTER_HUNT			0x10
#define WR3_AUTO_ENABLES		0x20
#define WR3_5BIT_RX				0x00
#define WR3_6BIT_RX				0x40
#define WR3_7BIT_RX				0x80
#define WR3_8BIT_RX				0xc0

#define WR4_PARITY_ENB			0x01
#define WR4_PARITY_ODD			0x00
#define WR4_PARITY_EVEN			0x02
#define WR4_SYNC_MODES			0x00
#define WR4_1STOPBIT			0x04
#define WR4_15STOPBIT			0x08
#define WR4_2STOPBIT			0x0c
#define WR4_8BITSYNC			0x00
#define WR4_16BITSYNC			0x10
#define WR4_SDLC_MODE			0x20
#define WR4_EXTSYNC				0x30
#define WR4_X1_CLK				0x00	/* 153600 bd, but... */
#define WR4_X16_CLK				0x40	/*   9600 bd */
#define WR4_X32_CLK				0x80	/*   4800 bd */
#define WR4_X64_CLK				0xc0	/*   2400 bd */

#define WR5_TXCRC_ENB			0x01
#define WR5_RTS					0x02
#define WR5_CCITT_CRC			0x00
#define WR5_CRC16_CRC			0x04
#define WR5_TX_ENB				0x08
#define WR5_SEND_BREAK			0x10
#define WR5_5BIT_TX				0x00
#define WR5_7BIT_TX				0x20
#define WR5_6BIT_TX				0x40
#define WR5_8BIT_TX				0x60
#define WR5_DTR					0x80

#define RR0_RCA					0x01
#define RR0_INT_PENDING			0x02
#define RR0_TBE					0x04
#define RR0_DCD					0x08
#define RR0_SYNC				0x10
#define RR0_CTS					0x20
#define RR0_UNDERRUN			0x40
#define RR0_BREAK_ABORT			0x80

#define RR1_ALL_SENT			0x01

#define RR1_PARER				0x10
#define RR1_ROVER				0x20
#define RR1_FRMERR				0x40
#define RR1_CRCERR				0x40
#define RR1_ENDFRAME			0x80

!----------------------------------------------------------------------
!
!	Output latch bits
!
#define O0_VOLUME 0x07
#define O0_INH    0x08
#define O0_AUDIOC 0x10
#define O0_CCIRC  0x20
#define O0_MTC    0x40
#define O0_MICM   0x80

#define O1_SRE    0x01
#define O1_SCE    0x02
#define O1_STE    0x04
#define O1_CLK    0x08
#define O1_SD     0x10
#define O1_RAS    0x20
#define O1_TPS    0x40
#define O1_TXOFF  0x80

!
!	Changes only by mainline.
!

#define O2_RA14   0x01
#define O2_RA15   0x02
#define O2_RS     0x04
#define O2_SMEM   0x08
#define O2_CS1    0x10
#define O2_CS2    0x20
#define O2_CLK    0x40
#define O2_DP     0x80

#define O2_XXX    O2_SMEM

#define O2_KEYPAD (O2_XXX | 0)
#define O2_LATCH  (O2_XXX | O2_CS1)
#define O2_LCD1   (O2_XXX | O2_CS2)
#define O2_LCD2   (O2_XXX | O2_CS2 | O2_CS1)

!----------------------------------------------------------------------
!
!	PIO and SIO controls
!
#define PA_CLK2   0x01
#define PA_HOOK   0x02
#define PA_WDR    0x04
#define PA_PWR    0x08
#define PA_CCIR   0xf0	/* all inputs */

#define PB_RAMA12 0x01	/* out */
#define PB_EXIN1  0x02
#define PB_EXIN2  0x04
#define PB_DCU    0x08
#define PB_RXOFF  0x10
#define PB_TMR0   0x20
#define PB_EXAL   0x40	/* out */
#define PB_PWROFF 0x80	/* out 3eh are inputs */

#define SA_KKINT  RR0_DCD
#define SA_DA     RR0_CTS

#define SB_PTT    RR0_CTS
#define SB_LOCAL  RR0_SYNC
#define SB_DMIDLE RR0_DCD	/* No MBUS activity */
#define SB_MON    WR5_DTR	/* Master ON pulse */

!	LATCH bits

#define BIT_AVAIL    0	/* not connected */
#define BIT_CALL     1
#define BIT_ROAM     2
#define BIT_KEYLIGHT 3
#define BIT_SERV     4
#define BIT_ON       5
#define BIT_LCDLIGHT 6
#define BIT_BRIGHT   7

!	Special LCD segments

#define SEG_V_U      0x43
#define SEG_V_D      0x53
#define SEG_PHONE    0x47
#define SEG_CLOCK    0x4b
#define SEG_COLON_UR 0x4f
#define SEG_COLON_UL 0x63
#define SEG_COLON_D  0x57
#define SEG_PHONE_NO 0x5b
#define SEG_MAST     0x5f
#define SEG_BAR_L    0x67
#define SEG_BAR_R    0x6f
#define SEG_STAR     0x6b
#define SEG_KEY      0x73
#define SEG_BOOK     0x77
#define SEG_CAR_D    0x7b
#define SEG_CAR_U    0x7f

!----------------------------------------------------------------------
!
!	realstart
!
	.text
	.org BASE

	di				! 1st op has to be 'di' for ram-loaded
	out [WD], a
	jp start

banner:
	.asciz BANNER

!----------------------------------------------------------------------
!
!	Interrupt table, short-aligned, on one page.
!
	.align 4

intvec:

sio_base:
	.word siob_tbe
	.word siob_esc
	.word siob_rca
	.word siob_src
	.word sioa_tbe
	.word sioa_esc
	.word sioa_rca
	.word sioa_src

pioa_base:
	.word pioa_int
piob_base:
	.word piob_int

#ifdef PROM
!----------------------------------------------------------------------
!
!	NMI
!
	.org 0x0066

	out [WD], a

	ld sp, 1f
	retn			! restart to
1:
	.word 0			! realstart in prom

#endif

!----------------------------------------------------------------------
!
!	Initialization commands for PIO, SIO and 8254
!	In that order, A/B, A/B, 0/1/2 !
!

init_chips:
1:
	.byte LO(pioa_base)	! load interrupt vector
	.byte 0xcf			! set operating mode 3
	.byte 0xff			! i/o selection, 1 for input, 0 for output
	.byte 0x97			! set interrupt control, IE, OR, LOW, MASK
	.byte 0xfe			! input bit monitoring mask, A0.
PIOA_INIT = . - 1b

1:
	.byte LO(piob_base)	! load interrupt vector
	.byte 0xcf			! set operating mode 3
	.byte 0x3e			! i/o selection, 1 for input, 0 for output
	.byte 0x97			! set interrupt control, IE, OR, LOW, MASK
	.byte 0xdf			! input bit monitoring mask, only B5 (TMR0)
PIOB_INIT = . - 1b

1:
	.byte WR0_CHANNEL_RESET
	.byte 4, WR4_1STOPBIT | WR4_X16_CLK
	.byte 3, WR3_8BIT_RX
	.byte 5, WR5_8BIT_TX | WR5_RTS | WR5_DTR
	.byte 1, WR1_STATUS_VECTOR | WR1_ESCINT_ENB
SIOA_INIT = . - 1b

1:
	.byte WR0_CHANNEL_RESET
	.byte 4, WR4_1STOPBIT | WR4_X16_CLK
	.byte 3, WR3_8BIT_RX | WR3_RX_ENB
	.byte 5, WR5_8BIT_TX | WR5_RTS | WR5_DTR | WR5_TX_ENB
	.byte 1, WR1_STATUS_VECTOR | WR1_ESCINT_ENB | WR1_TXINT_ENB | WR1_INT_ALLC
	.byte 2, LO(sio_base)
SIOB_INIT = . - 1b

!
!	All timer-gates are fixed on.
!	timer-clocks 0 and 1 are 4.032 MHz, 2 is 1968.75 Hz
!
!	System hz is 100, from PIO B5 and TMR0_2
!
MT_600HZ  = 4032000 / 600
MT_300HZ  = 4032000 / 300
MT_1750HZ = 4032000 / 1750
LP_3600HZ = 4032000 / 3600 / 100

	!  Control words, then initial counts.
1:
	.byte TMR_0 | TMR_BOTH | TMR_SQWAVE
	.byte TMR_1 | TMR_BOTH | TMR_SQWAVE
	.byte TMR_2 | TMR_BOTH | TMR_RATEGEN
TMR_INIT = . - 1b

	.word LP_3600HZ		! LP Filter
	.word 4				! TX- and audible tones
	.word 20			! 100 Hz clock interrupt to PIO B5

!----------------------------------------------------------------------
!
!	Initialization before main
!

start:

	ld sp, 0			! stack down from end of ram (last 16kB)

	ld a, HI(intvec)
	ld iv, a
	im2

	!
	!  Output latches
	!

	ld a, 0x00			! volume setting at lowest level
	ld [output_0], a
	out [OUT0], a

	ld a, O1_TXOFF		! no tx, idle all
	ld [output_1], a
	out [OUT1], a

	!
	!  Prep RAM selects for max workspace
	!
	ld a, O2_XXX		! select ram 0, i hope, on P8N
	out [OUT2], a

	ld a, 1
	out [CSMEM], a		! select full 16k page of ram on P8E, nop on P8N

	ld a, 0
	ld [txon], a

	!
	!	Analog outputs
	!
#if 0
	ld a, 85			! medium RX control
	ld [rfc], a
#else
	ld a, [rfc]
#endif
	out [DA_RFC], a

#if 0
	ld a, 0				! minimal TX power 
	ld [txpwr], a
#endif

	out [WD], a

	!
	!  Disable NMT modem
	!

	ld a, MDM_XXX
	out [MDM + MDMCTRL], a

	!
	!  PIO, SIO and 8254
	!

	ld hl, init_chips

	ld bc, MKWORD(PIO + ACTRL, PIOA_INIT)
	otir
	out [WD], a

	ld bc, MKWORD(PIO + BCTRL, PIOB_INIT)
	otir
	out [WD], a

	! Set output bits, and check if restarted from powerdown nmi.

	ld a, 0
	out [PIO+BDATA], a	! B0, EXAL, OFF all zero

	in a, [PIO+ADATA]
	and PA_PWR
	jp nz, powerdown

	ld bc, MKWORD(SIO + ACTRL, SIOA_INIT)
	in a, [c]
	otir
	out [WD], a

	ld bc, MKWORD(SIO + BCTRL, SIOB_INIT)
	in a, [c]
	otir
	out [WD], a

	ld bc, MKWORD(TMR + TMRCTRL, TMR_INIT)
	otir
	ld bc, MKWORD(TMR + 0, 2)
	otir
	ld bc, MKWORD(TMR + 1, 2)
	otir
	ld bc, MKWORD(TMR + 2, 2)
	otir

	out [WD], a

	ld hl, mbusrx_buf
	ld [mbusrx_rp], hl
	ld [mbusrx_wp], hl

	ld hl, mbustx_buf
	ld [mbustx_rp], hl
	ld [mbustx_wp], hl

	out [WD], a

	ld de, _bss
	ld a, 0
1:
	out [WD], a
	ld hl, _end
	and a
	sbc hl, de
	jr z, 1f
	ld [de], a
	inc de
	jr 1b
1:
	call blip

	ld a, -1
	ld [key], a

	out [WD], a

#if 0
	!
	!  Initialize NMT modem
	!

	ld a, 0x04
	out [MDM + MDMCTRL], a		! Write 04 to control register

	ld hl, 10000
1:
	out [WD], a

	dec hl
	ld a, h
	or l
	jr nz, 1b					! Wait at least 1 bit.

	ld a, 0x00
	out [MDM + MDMCTRL], a		! Write zeros to control register

#endif

	ei

	jp main

!----------------------------------------------------------------------

doreti:
	pop af
	ei
	reti

!----------------------------------------------------------------------
!
!  SIO interrupts
!

siob_tbe:
	push af
	ld a, [mbustx_cnt]
	dec a
	ld [mbustx_cnt], a
	jr z, siob_tbe_close
	push hl
	ld hl, [mbustx_rp]
	ld a, [hl]
	inc l
	ld [mbustx_rp], hl
	pop hl
	out [SIO+BDATA], a
	ld a, 25
	ld [mbus_timer], a		! keep transmit timer on
	jp doreti

siob_tbe_close:
	ld a, WR0_RESET_TXINT
	out [SIO+BCTRL], a
	jp doreti

siob_rca_drop:
	call bleep

siob_rca_ignore:
	in a, [SIO+BDATA]

	jp doreti

siob_rca:
	push af
	ld a, [mbus_timer]
	or a
	jr nz, siob_rca_ignore
	ld a, [mbusrx_cnt]
	inc a
	jr z, siob_rca_drop			! no room
	ld [mbusrx_cnt], a
	push hl
	ld hl, [mbusrx_wp]
	in a, [SIO+BDATA]
	ld [hl], a
	inc l
	ld [mbusrx_wp], hl
	pop hl
	jp doreti

!
!	SIO B ext/status change
!
siob_esc:
	push af
	in a, [SIO+BCTRL]
	and SB_PTT
	ld [pttdn], a
	ld a, WR0_RESET_ESCINT
	out [SIO+BCTRL], a
	jp doreti

!
!	SIO B special receive condition
!
siob_src:
	push af
	in a, [SIO+BDATA]
	ld a, WR0_ERROR_RESET
	out [SIO+BCTRL], a

	call bleep

	jp doreti

sioa_tbe:
	push af
	ld a, WR0_RESET_TXINT
	out [SIO+ACTRL], a
	jp doreti

sioa_rca:
	push af
	in a, [SIO+ADATA]
	jp doreti

!
!	SIO A ext/status change
!
sioa_esc:
	push af

	in a, [SIO+ACTRL]
	and SA_DA
	ld a, 0
	jp nz, 1f
	ld a, [keydown]
	or a
	jr nz, 2f
	ld a, 1
	ld [keydown], a
	jr 2f
1:
	ld a, 0
	ld [key_timer], a
	ld [keydown], a
2:

#if 0
	in a, [MDM + MDMCTRL]
	and MDM_RXRDY
	jr z, 1f
	in a, [MDM + MDMDATA]
	call iputc
1:
#endif

	ld a, WR0_RESET_ESCINT
	out [SIO+ACTRL], a
	jp doreti

!
!	SIO A special receive condition
!
sioa_src:
	push af
	in a, [SIO+ADATA]
	ld a, WR0_ERROR_RESET
	out [SIO+ACTRL], a
	jp doreti

!----------------------------------------------------------------------
!
!  PIO interrupts
!

!	Change in B5

piob_int:
	push af

	in a, [TMR+TMRCTRL]		! dummy read to raise PIO B5 (remove int)
	jp doreti

!	Change in A0, 1968.75 Hz square

pioa_int:
	ex af
	exx
	dec c
	jr z, 1f
	ex af
	exx
	ei
	reti
1:
	out [WD], a				! keep watchdog happy.

	ld a, [mt_timer]
	or a
	jr z, 1f
	dec a
	ld [mt_timer], a
	jr nz, 1f
	ld a, [output_0]
	and ~O0_MTC
	ld [output_0], a
	out [OUT0], a
1:

	!
	!  Decrement timer for handset lights
	!
	ld a, [keydown]
	or a
	jr nz, 1f				! not if button still pressed
	ld hl, [light_ltimer]
	ld a, h
	or l
	jr z, 1f				! not if timer not running
	dec hl
	ld [light_ltimer], hl
1:
	!
	!  Decrement MBUS activity timer
	!
	ld a, [mbus_timer]
	or a
	jr z, 1f
	dec a
	ld [mbus_timer], a
1:
	!
	!  See if keypad has data coming and debouncing
	!  allows it to be fetched
	!
	ld a, [keydown]
	or a
	jr z, 1f			! not key down
	inc a
	jr z, 1f			! been a long time, 256 softicks
	ld [keydown], a
	cp 10
	jr nz, 1f			! not at debounce-position
	ld a, [sir]
	set KEYSIR, a
	ld [sir], a			! go get it.
1:

	!	S meter and squelch control

#if 1
	ld a, [txon]
	or a
	jr nz, 9f
#endif

	in a, [AD]
	ld b, a					! current rssi
	ld [rssi], a

	ld a, 0
	out [AD_RSSI], a		! value not important
	out [AD_RSSI], a		! twice to settle analog mux in case it changed

#if 0
	ld a, [squelch_open]
	or a
	jr z, 1f				! go if closed
#endif
	ld a, [squelch_forced]
	or a
	jr nz, 2f

	ld a, [squelch_limit_close]
	cp b					! close_limit - rssi
	jr c, 1f

	ld a, 0
	ld [squelch_open], a
	ld a, [output_0]
	and ~O0_AUDIOC
	ld [output_0], a
	out [OUT0], a
1:
	ld a, [squelch_limit_open]
	cp b					! open_limit - rssi
	jr nc, 1f
2:
	ld a, 1
	ld [squelch_open], a
	ld a, [output_0]
	or O0_AUDIOC
	ld [output_0], a
	out [OUT0], a
1:

	!  smoothing rssi display

	ld a, [srssi]
	cp b					! smoothed - rssi
	jr c, 1f				! go if raising
	ld a, [rssi_timer]
	dec a
	jr nz, 2f				! go if still timing
1:
	ld a, b
	ld [srssi], a
	ld a, 100
2:
	ld [rssi_timer], a

9:

	!  Typematic

	ld a, [key_timer]
	or a
	jr z, 9f
	dec a
	ld [key_timer], a
	jr nz, 9f
	ld a, [key_speed]
	ld [key_timer], a
	dec a
	jr z, 1f
	ld [key_speed], a
	call blip
1:
	ld a, [lastkey]
	ld [key], a

9:
	ld c, 1969 / 100	! hard ticks to next
	exx

	!  softlevel

	ld a, [sir]
	or a
	jr nz, 2f		! there are sirs
1:
	ex af
8:
	ei
	reti
2:
	bit INSIR, a
	jr nz, 1b		! already here

	set INSIR, a
	ld [sir], a		! note sir starting

	ex af

	push af
	push bc
	push de
	push hl
	push ix
	push iy			! sir runs with normal regset

	call 8b
	call dosir
	di

	ld a, [sir]
	res INSIR, a
	ld [sir], a

	pop iy
	pop ix
	pop hl
	pop de
	pop bc
	pop af

	ei
	ret

dosir:

	ld a, [sir]
	bit KEYSIR, a
	jr z, 1f
	res KEYSIR, a
	ld [sir], a
	call keypad
1:
	ld a, [sir]
	bit DPYSIR, a
	jr z, 1f
	res DPYSIR, a
	ld [sir], a
	call display
1:
	ret

segtest:
	.byte SEG_V_U, SEG_V_D, SEG_PHONE, SEG_CLOCK
	.byte SEG_COLON_UR, SEG_COLON_UL, SEG_COLON_D, SEG_PHONE_NO
	.byte SEG_MAST, SEG_BAR_L, SEG_BAR_R, SEG_STAR
	.byte SEG_KEY, SEG_BOOK, SEG_CAR_D, SEG_CAR_U
	.byte 0xFF

!----------------------------------------------------------------------
!
!	Mainline
!

main:

	ld c, 0xa
	call putchar

	ld de, banner
1:
	ld a, [de]
	inc de
	or a
	jr z, 1f
	ld c, a
	call putchar
	jr 1b
1:
	ld de, segtest
1:
	ld a, [de]
	inc de
	inc a
	jr z, 1f
	dec a
	call segment_on
	jr 1b
1:

	ld hl, 100 * 15
	ld [light_ltimer], hl

	ld hl, sir
	set DPYSIR, [hl]

	!
	!	XXX
	!

#if 0
	ld a, 0
	ld [volume], a

	ld a, 0x40
	ld [squelch_limit_open], a
#endif
	ld a, [squelch_limit_open]
	sub 8
	ld [squelch_limit_close], a

	ld hl, 1024			! 12.5 step
	ld [bstep_R], hl

#if 0
	!  Enable modem receiver

	ld a, MDM_RXENB | MDM_RXFMT
	out [MDM + MDMCTRL], a
#endif

	!
	!  End XXX
	!
	call set_vola

	ld a, 0
	ld [batt], a
	ld [rssi], a

	call reset_synth

	call redraw

	ld hl, segments
	ld b, SIZE(segments)
	ld a, 0
1:
	ld [hl], a
	inc hl
	djnz 1b

1:
	ld a, [keydown]
	or a
	jp nz, 1b			! wait release

	ld c, 0xa
	call putchar

#if 0
	imm_ahl(433500)
	call freq2div
	save_ahl(tx_divisor)
#else
	load_ahl(tx_divisor)
#endif
	call change_synth

mainloop:
1:
	call txset
	call redraw

	ld a, [key]
	cp -1
	jp z, 1b

	ld c, a
	ld a, -1
	ld [key], a
	ld a, c

	call dokey
	jp mainloop

dokey:				! whatever we call will do ret

	cp 10
	jp c, insdig

	cp 'C'
	jp z, backspace
	cp 'S'
	jp z, save
	cp 'E'
	jp z, execute
	cp '#'
	jp z, command
	cp 'R'
	jp z, toggle_rptr
	cp '*'
	jp z, beep1750
	cp '+'
	jp z, up
	cp '-'
	jp z, dn
	cp 'B'
	jp z, monitor_audio

	ret

next_shift_state:
	ld a, [shift_state]
	inc a
	cp SHIFT_STATES
	jr c, 1f
	ld a, 0
1:
	ld [shift_state], a
	ret

dn:
	ld a, [shift_state]
	sla a
	sla a
	jr 1f
up:
	ld a, [shift_state]
	sla a
	sla a
	inc a
	inc a
1:
	ld b, 0
	ld c, a
	ld hl, 1f
	add hl, bc
	ld a, [hl]
	inc hl
	ld h, [hl]
	ld l, a
	jp [hl]
1:
	.word dn_vola, up_vola
	.word dn_sqlv, up_sqlv
	.word dn_memo, up_memo
	.word dn_freq, up_freq

	!.word dn_pwr,  up_pwr
	!.word dn_rfc,  up_rfc

	SHIFT_STATES = (. - 1b) / (2 * 2)

beep1750:

	call tx_on
	di
	ld a, LO(MT_1750HZ)
	out [TMR + 1], a
	ld a, HI(MT_1750HZ)
	out [TMR + 1], a

	ld a, [output_0]
	or O0_CCIRC | O0_MICM
	ld [output_0], a
	out [OUT0], a
	ei
1:
	call redraw
	ld a, [keydown]
	or a
	jp nz, 1b

	di
	ld a, [output_0]
	and ~(O0_CCIRC | O0_MICM)
	ld [output_0], a
	out [OUT0], a
	ei

	call tx_off

	ret

toggle_rptr:
	load_ahl(tx_sub)
	or h
	or l
	jr z, 1f
	imm_ahl(0)
	save_ahl(tx_sub)
	ret
1:
	!
	!  If current frequency below 300MHz, use 600 kHz difference
	!  Otherwise 1600 kHz
	!
	load_ahl(tx_divisor)
	call div2freq
	cp 300000 / 65536		! Just compare _A_HL
	imm_ahl(1600)
	jr nc, 1f
	imm_ahl(600)
1:
	call freq2div
	save_ahl(tx_sub)
	ret

command:
	ld a, [digidx]
	or a
	ret z
	cp 1
	jr nz, 1f
	call a2i
	ld a, l
	and 3
	ld [txpwr], a
	ret
1:
	cp 3
	jr nz, 1f
	call a2i
	ld a, l
	ld [rfc], a
	out [DA_RFC], a
	ret
1:
	cp 4
	jr nz, 1f
	call a2i
	ld [bstep_R], hl
	call reset_synth
	ret
1:
	call a2i
	save_ahl(tx_divisor)
	ret

execute:
	ld a, [digidx]
	or a
	jp z, next_shift_state	! 0 digits
	cp 3
	jp c, gomem				! 1...2 digits
	cp 5
	jr nc, gofreq			! 5... digits

	cp 3
	jr z, 1f
	call a2i
	ld de, 430000 % 65536
	ld b,  430000 / 65536
	jr 2f
1:
	call a2i
	ld de, 433000 % 65536
	ld b,  433000 / 65536
2:
	add hl, de
	adc b
	jr 2f

gofreq:
	call a2i
2:
	save_ahl(currfreq)
	call freq2div
	save_ahl(tx_divisor)
	call change_synth

	ret

#define PACKMEM(f) .byte 0 ; .word (f) % 65536 ; .byte (f) / 65536

save:
	ld a, [digidx]
	or a
	jr nz, 1f
	ld a, [currmem]
	jr 2f
1:
	call a2i
	ld a, l
2:
	and 127
	ld [currmem], a
	ld b, 0
	ld c, a
	ld hl, memories
	add hl, bc
	add hl, bc
	add hl, bc
	add hl, bc
	push hl
	pop ix

	load_ahl(currfreq)
	ld [ix + 1], l
	ld [ix + 2], h
	ld [ix + 3], a

	ret

up_memo:
	ld a, 1
	jr 2f

dn_memo:
	ld a, -1
2:
	ld hl, currmem
	add [hl]		! -1 -> 0/-2, 99 -> 100/98

	cp -1
	jr z, 1f
	cp -2
	jr nz, 2f
1:
	ld a, 99
2:
	cp 100
	jr c, 2f
	ld a, 0
2:
	jp gomem_a

gomem:
	call a2i
	ld a, l
	and 127

gomem_a:

	ld [currmem], a
	ld b, 0
	ld c, a
	ld hl, memories
	add hl, bc
	add hl, bc
	add hl, bc
	add hl, bc
	push hl
	pop ix

	ld l, [ix + 1]
	ld h, [ix + 2]
	ld a, [ix + 3]
	save_ahl(currfreq)
	call freq2div
	save_ahl(tx_divisor)
	call change_synth

	ret

insdig:
	ld e, a

	ld a, [digidx]
	cp 16
	jr nc, 1f
	ld hl, digbuf
	ld b, 0
	ld c, a
	add hl, bc
	inc a
	ld [hl], e
	ld [digidx], a
1:
	ret

backspace:
	ld a, [digidx]
	or a
	jr z, 1f
	sub 1
	adc 0
	ld [digidx], a
	ret
1:
	ld a, 0
	ld [shift_state], a
	ret

monitor_audio:
	ld a, 1
	ld [squelch_forced], a
1:
	call redraw
	ld a, [keydown]
	or a
	jr nz, 1b
	ld a, 0
	ld [squelch_forced], a
	ret

restart_prom:
1:
	ld a, [keydown]
	or a
	jp nz, 1b			! wait release

	jp 0x0000
	
up_sqlv:
	ld a, 1
	jr 1f
dn_sqlv:
	ld a, -1
1:
	push af
	ld hl, squelch_limit_open
	add [hl]
	ld [hl], a
	pop af
	ld hl, squelch_limit_close
	add [hl]
	ld [hl], a
	ret

up_rfc:
	ld a, 1
	jr 1f
dn_rfc:
	ld a, -1
1:
	ld hl, rfc
	add [hl]
	ld [hl], a
	out [DA_RFC], a
	ret

set_vola:
	ld a, 0
	jr 1f
up_vola:
	ld a, 1
	jr 1f
dn_vola:
	ld a, -1
1:
	ld hl, volume
	add [hl]

	cp 10
	jr z, 1f		! 9 -> 10
	jr c, 2f		! 0 -> -1
	ld a, 0
	jr 2f
1:
	ld a, 9
2:
	ld [hl], a
	cp 0
	jr z, 1f
	cp 1
	jr z, 2f
	sub 2				! 9 - 2 -> 7
	and 7
	ld b, a
	di
	ld a, [output_0]
	and ~O0_VOLUME
	and ~O0_INH
	or b
	ld [output_0], a
	out [OUT0], a
	ei
	ret
1:
	di
	ld a, [output_0]
	and ~O0_VOLUME
	or O0_INH
	ld [output_0], a
	out [OUT0], a
	ei
	ret
2:
	di
	ld a, [output_0]
	and ~O0_VOLUME
	and ~O0_INH
	ld [output_0], a
	out [OUT0], a
	ei
	ret

up_freq:
	load_ahl(tx_divisor)
	ld bc, 1
	add hl, bc
	adc 0
	and 1						! only 17-bit divisor
	save_ahl(tx_divisor)
	call div2freq
	save_ahl(currfreq)
	call change_synth
	ret

dn_freq:
	load_ahl(tx_divisor)
	ld bc, 0
	scf
	sbc hl, bc
	sbc 0
	and 1						! only 17-bit divisor
	save_ahl(tx_divisor)
	call div2freq
	save_ahl(currfreq)
	call change_synth
	ret

txset:
	ld a, [pttdn]
	or a
	ret z

	call tx_on
1:
	call redraw
	ld a, [pttdn]
	or a
	jp nz, 1b

	call tx_off
	ret

!----------------------------------------------------------------------
!
!	Powerdown
!
powerdown:
	ld a, PB_PWROFF
	out [PIO+BDATA], a
	halt
	jp powerdown

!----------------------------------------------------------------------
!
!	MBUS putchar, character in c
!

putchar:
	ld hl, mbustx_cnt
putchar_1:
	ld a, [hl]
	inc a
	jr z, putchar_1			! wait until not full
	di
	ld a, [hl]
	inc [hl]
	or a
	jr z, putchar_start		! tx not active, go start it
	ld hl, [mbustx_wp]
	ld [hl], c
	inc l
	ld [mbustx_wp], hl
	ei
	ret
putchar_start:
	ld a, c
	out [SIO+BDATA], a
	ld a, 25
	ld [mbus_timer], a
putchar_out:
	ei
	ret

iputc:
	push af
	ld a, [mbustx_cnt]
	inc a
	jr z, 1f				! full
	ld [mbustx_cnt], a
	dec a
	jr z, 2f				! tx not active, go start it
	pop af
	push hl
	ld hl, [mbustx_wp]
	ld [hl], a
	inc l
	ld [mbustx_wp], hl
	pop hl
	ret
1:
	pop af
	ret
2:
	pop af
	out [SIO+BDATA], a
	ld a, 25
	ld [mbus_timer], a
	ret


getchar:
	ld hl, mbusrx_cnt
	ld a, [hl]
	or a
	jr z, getchar
	di
	dec [hl]
	ld hl, [mbusrx_rp]
	ld a, [hl]
	inc l
	ld [mbusrx_rp], hl
	ei
	ret

	.align 8

keytbl:
	.byte 'Z','Z','Z','Z','Z','Z','Z','Z'
	.byte 'Z','Z','Z','Z','+','-','?','B'
	.byte 'E','*', 0 ,'#','R', 7 , 8 , 9
	.byte 'S', 4 , 5 , 6 ,'C', 1 , 2 , 3

keypad:

	call blip

	ld hl, 100 * 15
	ld [light_ltimer], hl

	!
	!	strobe the data into shifter
	!
	ld a, O2_KEYPAD
	out [OUT2], a
	ld a, O2_KEYPAD | O2_CLK
	out [OUT2], a
	ld a, O2_KEYPAD
	out [OUT2], a
	ld a, O2_LCD1
	out [OUT2], a

	in a, [PIO+BDATA]		! from LDR
	and PB_DCU
	ld [dark], a

	!
	!	shift 5 keycode bits to l
	!
	ld l, 0		! collect bits here
	ld h, 5		! this many
1:
	ld a, O2_LCD1 | O2_CLK
	out [OUT2], a
	ld a, O2_LCD1
	out [OUT2], a
	in a, [PIO+BDATA]
	and PB_DCU
	sub 1
	rl l

	dec h
	jr nz, 1b

	ld h, HI(keytbl)

	ld a, [hl]
	ld [key], a

	cp '+'
	jr z, 1f
	cp '-'
	jr z, 1f
	jr 2f
1:
	ld [lastkey], a
	ld a, 50
	ld [key_timer], a
	ld a, 25
	ld [key_speed], a
2:

	ret

bleep:
	ld a, LO(MT_300HZ)
	out [TMR + 1], a
	ld a, HI(MT_300HZ)
	out [TMR + 1], a

	ld a, [output_0]
	or O0_MTC
	ld [output_0], a
	out [OUT0], a

	ld a, 100
	ld [mt_timer], a

	ret

blip:
	ld a, LO(MT_600HZ)
	out [TMR + 1], a
	ld a, HI(MT_600HZ)
	out [TMR + 1], a

	ld a, [output_0]
	or O0_MTC
	ld [output_0], a
	out [OUT0], a

	ld a, 2
	ld [mt_timer], a

	ret

!  0 ->  8 -> 0
!  1 ->  9 -> 0
!  2 -> 10 -> 0
!  7 -> 15 -> 0
!  8 -> 16 -> 1
!  9 -> 17 -> 1

dpysig:
	add 8
	jr nc, 1f
	ld a, 0xff
1:
	srl a
	srl a
	srl a
	srl a
	jp dpydig

dpyval99:

	ld c, -1
1:
	inc c
	sub 10
	jr nc, 1b
	add 10
	push af

	ld a, c
	or a
	jr nz, 1f
	ld a, ' '
1:
	call dpydig
	pop af
	call dpydig
	ret

dpyhex:
	push af
	srl a
	srl a
	srl a
	srl a
	call 1f
	pop af
1:
	and 0xf
	cp 10
	jr c, 1f
	add 'A' - 10 - '0'
1:
	add '0'
	call dpydig
	ret

!----------------------------------------------------------------------
!
!	Build up display buffer
!

redraw:

	!	LEDs

	ld b, MASK(BIT_ON)

	ld a, [dark]
	or a
	jr nz, 1f
	set BIT_BRIGHT, b
1:
	ld a, [squelch_open]
	or a
	jr z, 1f
	set BIT_SERV, b
1:
	ld a, [txon]
	or a
	jr z, 1f
	set BIT_ROAM, b
1:
	ld hl, [light_ltimer]
	ld a, h
	or l
	jr z, 1f
	set BIT_LCDLIGHT, b
	set BIT_KEYLIGHT, b
1:
	ld hl, indicators
	ld [hl], b

	!  Cursor colons

	ld a, [shift_state]
	sub 1					! carry if 0
	push af
	ld a, SEG_COLON_UL
	call segment_onoff

	pop af
	sub 1					! carry if initially 1
	push af
	ld a, SEG_COLON_UR
	call segment_onoff

	ld a, [digidx]			! if there are digits, clean bottom row
	or a
	jr z, 1f
	pop af
	ld a, SEG_COLON_D
	and a
	call segment_onoff
	ld a, SEG_V_U
	and a
	call segment_onoff
	ld a, SEG_V_D
	and a
	call segment_onoff
	jr 2f
1:
	pop af
	sub 1					! carry if initially 2
	push af
	ld a, SEG_COLON_D
	call segment_onoff

	pop af
	sub 1					! carry if initially 3
	push af
	ld a, SEG_V_U
	call segment_onoff
	pop af
	ld a, SEG_V_D
	call segment_onoff
2:
	!  Icons, repeater, power, battery

	load_ahl(tx_sub)
	or h
	or l
	sub 1
	ccf
	ld a, SEG_MAST    
	call segment_onoff

	ld a, [txpwr]
	rr a
	push af
	ld a, SEG_CAR_D
	call segment_onoff
	pop af
	rr a
	ld a, SEG_CAR_U
	call segment_onoff

	!  Upper row, volume squelch srssi


#if 0
ld de, segs_u_digit_5
ld a, [tx_divisor + 2]
call dpyhex
ld a, [tx_divisor + 1]
call dpyhex
ld a, [tx_divisor + 0]
call dpyhex

#else

	ld de, segs_u_digit_4

	ld a, [volume]
	call dpydig					! volume

	ld a, [squelch_limit_open]
	call dpydiv13					! squelch setting

	ld a, [srssi]
	call dpydiv13					! srssi

#endif

	!  Lower left, 2 digits, lower right, 8 digits

	ld de, segs_dl_digit_1

	ld a, [digidx]		! how many digits
	or a
	jr nz, 9f

	ld a, [currmem]
	cp 100
	jr c, 1f
	ld a, '-'
	call dpydig
	ld a, '-'
	call dpydig
	jr 2f
1:
	call dpyval99
2:
	ld a, ' '
	call dpydig

	call div2dpy		! show frequency
	jr 8f
9:
	ld hl, digbuf
	cp 10				! positions max
	jr c, 1f
	ld a, 10
1:
	ld b, a				! to show
	neg
	add 10
	ld c, a				! to blank
1:
	ld a, [hl]
	inc hl
	push hl
	call dpydig
	pop hl
	djnz 1b
	ld a, c
	or a
	jr z, 9f
	ld b, a
1:
	ld a, ' '
	call dpydig
	djnz 1b
8:

	ld hl, sir
	set DPYSIR, [hl]

	ret

display:

	ld hl, segments

	ld a, O2_LCD2
	out [OUT2], a

	ld a, O2_LCD1		! start 1st half of LCD1
	call display_group
	call display_bit_zero

	ld a, O2_LCD2		! start 1st half of LCD2
	call display_group
	call display_bit_zero

	ld a, O2_LCD1		! start 2nd half of LCD1
	call display_group
	call display_bit_one

	ld a, O2_LCD2		! start 2nd half of LCD2
	call display_group
	call display_bit_one

	ld a, O2_LCD1		! unselect LATCH
	out [OUT2], a

	call display_byte

	ld a, O2_LATCH		! select LATCH
	out [OUT2], a
	ld a, O2_LCD1		! unselect LATCH
	out [OUT2], a

	ret

display_group:
	out [OUT2], a

	call display_bit_zero
	call display_byte
	call display_byte
	call display_byte
	call display_byte
	ret

	!
	!	4032000 Hz pclk
	!	0.24 us T-state
	!
display_byte:
	ld c, [hl]
	inc hl
	ld b, 8
1:
	srl c
	call display_bit
	djnz 1b
	ret

display_bit:
	jr nc, display_bit_zero
display_bit_one:
	or O2_DP
display_bit_zero:
	or O2_CLK
	out [OUT2], a		! data & clock set	11 T	2.7 us
	and ~O2_CLK
	out [OUT2], a		! clock edge		2.7 us
	and ~O2_DP
	out [OUT2], a		! clear data		2.7 us

	ret

puthex:
	push bc
	ld a, c
	srl a
	srl a
	srl a
	srl a
	call 1f
	pop bc
	ld a, c
1:
	and 0xf
	cp 10
	jr c, 1f
	add 'A' - 10 - '0'
1:
	add '0'
	ld c, a
	call putchar
	ret

bitmasks:
	.byte 1, 2, 4, 8, 0x10, 0x20, 0x40, 0x80

segment_onoff:
	jr nc, segment_off

segment_on:

	ld h, HI(segments)	; ASSERT(LO(segments) == 0)
	ld l, a
	srl l
	srl l
	srl l				! byte ptr ok

	!	Uhhuh. Maybe not so bad as it looks :)

	and 7				! bit number

	cp 4; jr nc, 4f			! >= 4
	cp 2; jr nc, 2f			! >= 2, < 4
	cp 1; jr nc, 1f			! >= 1, < 2
	set 0, [hl] ; ret
1:	set 1, [hl] ; ret
2:	cp 3; jr nc, 3f			! >= 3, < 4
	set 2, [hl] ; ret
3:	set 3, [hl] ; ret
4:	cp 6; jr nc, 6f			! >= 6
	cp 5; jr nc, 5f			! >= 5, < 6
	set 4, [hl] ; ret
5:	set 5, [hl] ; ret
6:	cp 7; jr nc, 7f			! >= 7
	set 6, [hl] ; ret
7:	set 7, [hl] ; ret

segment_off:

	ld h, HI(segments)	; ASSERT(LO(segments) == 0)
	ld l, a
	srl l
	srl l
	srl l				! byte ptr ok

	!	Uhhuh. Maybe not so bad as it looks :)

	and 7				! bit number

	cp 4; jr nc, 4f			! >= 4
	cp 2; jr nc, 2f			! >= 2, < 4
	cp 1; jr nc, 1f			! >= 1, < 2
	res 0, [hl] ; ret
1:	res 1, [hl] ; ret
2:	cp 3; jr nc, 3f			! >= 3, < 4
	res 2, [hl] ; ret
3:	res 3, [hl] ; ret
4:	cp 6; jr nc, 6f			! >= 6
	cp 5; jr nc, 5f			! >= 5, < 6
	res 4, [hl] ; ret
5:	res 5, [hl] ; ret
6:	cp 7; jr nc, 7f			! >= 7
	res 6, [hl] ; ret
7:	res 7, [hl] ; ret


!
!	Calculate R and misc stuff
!
!	25 kHz
!	 512, 12800 /  512  quot 25 rem 0, /= 25 shift 0
!
!	12.5 kHz
!	1024, 12800 / 1024  quot 12 rem 512
!	      12800 /  512  quot 25 rem 0, /= 25 shift 1
!
!	6.25 kHz
!	2048, 12800 / 2048  quot  6 rem 512
!	      12800 / 1024  quot 12 rem 512
!	      12800 /  512  quot 25 rem 0, /= 25 shift 2
!
!	3.125 kHz
!	4096, 12800 / 4096  quot  3 rem 512
!	      12800 / 2048  quot  6 rem 512
!	      12800 / 1024  quot 12 rem 512
!	      12800 /  512  quot 25 rem 0, /= 25 shift 3
!
!	10 kHz
!	1280, 12800 / 1280  quot 10 rem 0, /= 10 shift 0
!
!	5 kHz
!	2560, 12800 / 2560  quot  5 rem 0, 5 too small, use /= 10 shift 1
!
!

reset_synth:

	ld de, [bstep_R]
	ld b, 0			! try with no shift, will be 0..7
2:
	ld hl, 12800
	ld c, 0
1:
	inc c			! increment x2y-divider, will be from 1..
	jr z, 9f		! oops
	and a
	sbc hl, de
	jr z, 1f		! exact !
	jr nc, 1b		! division not complete

	inc b			! try with more shift.
	jr z, 9f		! oops
	srl d
	rr e			! try divide with smaller value
	ld a, e
	or d
	jr z, 9f		! oops
	jr 2b
1:
	ld a, c
	cp 10
	jr nc, 2f
	inc b			! hack in 5kHz on 70cm by changing 5/0 to 10/1
	sla c
	jr nz, 1b
9:
	ld b, 1
	ld c, 25
	ld de, 1024
	ld [bstep_R], de
2:
	ld [bstep_cfg], bc

	imm_ahl(86512)
	call freq2div
	save_ahl(rx_sub)

	load_ahl(currfreq)
	call freq2div
	save_ahl(tx_divisor)

	call toggle_rptr
	call toggle_rptr

	call synth_setup
	call change_synth

	ret

!
!	Configure RX and TX synths
!

txmode_synth:

	ld c, 0xf8		! Max deviation (b7..b4), TX synth operational (b3,b2)
	jp 9f

synth_setup:

	ld c, 0xf4		! Max deviation (b7..b4), TX synth OFF (b3,b2)
9:
	ld a, O1_TXOFF
	out [OUT1], a

	ld b, 8
	call send_to_synth

	or O1_SCE
	call strobe_to_synth

	ld hl, [bstep_R]	! channel spacing
	ld c, h
	ld b, 8			! has excess bits... they overflow ok
	call send_to_synth
	ld c, l
	ld b, 8
	call send_to_synth
	ld c, 0x80		! single "1" bit
	ld b, 1
	call send_to_synth	! to R register

	or O1_SRE | O1_STE
	call strobe_to_synth	! both R same

	ret

change_synth:

	load_ahl(tx_divisor)
	ld e, a

	ld a, O1_TXOFF
	out [OUT1], a

	ld c, e
	rrc c
	ld b, 1
	call send_to_synth		! send MSb

	ld c, h
	ld b, 8
	call send_to_synth
	ld c, l
	ld b, 8
	call send_to_synth
	ld c, 0x00			! single "0" bit
	ld b, 1
	call send_to_synth	! to A/N registers

	or O1_STE
	call strobe_to_synth

	load_abc(rx_sub)
	add hl, bc
	adc e				! EHL + ABC
	ld e, a

	ld a, O1_TXOFF
	out [OUT1], a

	ld c, e
	rrc c
	ld b, 1
	call send_to_synth

	ld c, h
	ld b, 8
	call send_to_synth
	ld c, l
	ld b, 8
	call send_to_synth
	ld c, 0x00			! single "0" bit
	ld b, 1
	call send_to_synth	! to A/N registers

	or O1_SRE
	call strobe_to_synth

	ret

strobe_to_synth:

	out [OUT1], a
	and ~(O1_SCE | O1_STE | O1_SRE)
	out [OUT1], a
	ret

send_to_synth:
2:
	rl c
	jr nc, 1f
	or O1_SD
1:
	out [OUT1], a
	or O1_CLK
	out [OUT1], a
	and ~(O1_CLK)
	out [OUT1], a
	and ~(O1_CLK | O1_SD)
	out [OUT1], a
	djnz 2b

	ret

!----------------------------------------------------------------------
!
!     XXYYZZ
!   9876543210
!
!      V P S        vola pwr srssi
!   CH  433500    channel  frequency
!

!	down left digit-area

segs_dl_digit_1:	.byte 56, 57, 58, 59,	120, 121, 122
segs_dl_digit_0:	.byte 60, 61, 62, 63,	124, 125, 126


!	down right digit-area

segs_dr_digit_7:	.byte  0,  1,  2,  3,	 64,  65,  66
segs_dr_digit_6:	.byte  4,  5,  6,  7,	 68,  69,  70
segs_dr_digit_5:	.byte  8,  9, 10, 11,	 72,  73,  74
segs_dr_digit_4:	.byte 12, 13, 14, 15,	 76,  77,  78
segs_dr_digit_3:	.byte 16, 17, 18, 19,	 80,  81,  82
segs_dr_digit_2:	.byte 20, 21, 22, 23,	 84,  85,  86
segs_dr_digit_1:	.byte 24, 25, 26, 27,	 88,  89,  90
segs_dr_digit_0:	.byte 28, 29, 30, 31,	 92,  93,  94

!	upper digit-area

segs_u_digit_5:		.byte 32, 33, 34, 35,	 96,  97,  98
segs_u_digit_4:		.byte 36, 37, 38, 39,	100, 101, 102
segs_u_digit_3:		.byte 40, 41, 42, 43,	104, 105, 106
segs_u_digit_2:		.byte 44, 45, 46, 47,	108, 109, 110
segs_u_digit_1:		.byte 48, 49, 50, 51,	112, 113, 114
segs_u_digit_0:		.byte 52, 53, 54, 55,	116, 117, 118

!
!   -     4
!  | |  0   3
!   -     5
!  | |  1   2
!   -     6

	.align 8

font:

	.byte	0x5f	! 0 	0 1 2 3   4 - 6
	.byte	0x0c	! 1 	- - 2 3   - - -
	.byte	0x7a	! 2 	- 1 - 3   4 5 6
	.byte	0x7c	! 3 	- - 2 3   4 5 6
	.byte	0x2d	! 4 	0 - 2 3   - 5 -
	.byte	0x75	! 5 	0 - 2 -   4 5 6
	.byte	0x77	! 6 	0 1 2 -   4 5 6
	.byte	0x1c	! 7 	- - 2 3   4 - -
	.byte	0x7f	! 8 	0 1 2 3   4 5 6
	.byte	0x7d	! 9 	0 - 2 3   4 5 6
	.byte	0x3f	! A 	0 1 2 3   4 5 -
	.byte	0x67	! b 	0 1 2 -   - 5 6
	.byte	0x53	! C 	0 1 - -   4 - 6
	.byte	0x6e	! d 	- 1 2 3   - 5 6
	.byte	0x73	! E 	0 1 - -   4 5 6
	.byte	0x33	! F 	0 1 - -   4 5 -

	.byte	0x57	! G 	0 1 2 -   4 - 6

	.org font + ' '

	.byte	0x00	!   	- - - -   - - -
	.byte	0x0c	! ! 	- - 2 3   - - -
	.byte	0x09	! " 	0 - - 3   - - -
	.byte	0x66	! # 	- 1 2 -   - 5 6
	.byte	0x25	! $ 	0 - 2 -   - 5 -
	.byte	0x05	! % 	0 - 2 -   - - -
	.byte	0x6a	! & 	- 1 - 3   - 5 6
	.byte	0x08	! ' 	- - - 3   - - -
	.byte	0x0c	! ( 	- - 2 3   - - -
	.byte	0x03	! ) 	0 1 - -   - - -
	.byte	0x70	! * 	- - - -   4 5 6
	.byte	0x23	! + 	0 1 - -   - 5 -
	.byte	0x02	! , 	- 1 - -   - - -
	.byte	0x20	! - 	- - - -   - 5 -
	.byte	0x02	! . 	- 1 - -   - - -
	.byte	0x2a	! / 	- 1 - 3   - 5 -
	.byte	0x5f	! 0 	0 1 2 3   4 - 6
	.byte	0x0c	! 1 	- - 2 3   - - -
	.byte	0x7a	! 2 	- 1 - 3   4 5 6
	.byte	0x7c	! 3 	- - 2 3   4 5 6
	.byte	0x2d	! 4 	0 - 2 3   - 5 -
	.byte	0x75	! 5 	0 - 2 -   4 5 6
	.byte	0x77	! 6 	0 1 2 -   4 5 6
	.byte	0x1c	! 7 	- - 2 3   4 - -
	.byte	0x7f	! 8 	0 1 2 3   4 5 6
	.byte	0x7d	! 9 	0 - 2 3   4 5 6
	.byte	0x50	! : 	- - - -   4 - 6
	.byte	0x48	! ; 	- - - 3   - - 6
	.byte	0x2c	! < 	- - 2 3   - 5 -
	.byte	0x60	! = 	- - - -   - 5 6
	.byte	0x23	! > 	0 1 - -   - 5 -
	.byte	0x3a	! ? 	- 1 - 3   4 5 -
	.byte	0x7e	! @ 	- 1 2 3   4 5 6
	.byte	0x3f	! A 	0 1 2 3   4 5 -
	.byte	0x67	! b 	0 1 2 -   - 5 6
	.byte	0x53	! C 	0 1 - -   4 - 6
	.byte	0x6e	! d 	- 1 2 3   - 5 6
	.byte	0x73	! E 	0 1 - -   4 5 6
	.byte	0x33	! F 	0 1 - -   4 5 -
	.byte	0x57	! G 	0 1 2 -   4 - 6
	.byte	0x2f	! H		0 1 2 3   - 5 -
	.byte	0x0c	! I 	- - 2 3   - - -
	.byte	0x4e	! J 	- 1 2 3   - - 6
	.byte	0x2f	! K 	- 1 2 3   - - 6
	.byte	0x43	! L		0 1 - -   - - 6
	.byte	0x1f	! M		0 1 2 3   4 - -
	.byte	0x26	! N		- 1 2 -   - 5 -
	.byte	0x5f	! O 	0 1 2 3   4 - 6
	.byte	0x3b	! P 	0 1 - 3   4 5 -
	.byte	0x3d	! Q 	0 - 2 3   4 5 -
	.byte	0x1b	! R		0 1 - 3   4 - -
	.byte	0x75	! S 	0 - 2 -   4 5 6
	.byte	0x13	! T		0 1 - -   4 - -
	.byte	0x4f	! U 	0 1 2 3   - - 6
	.byte	0x4f	! V 	0 1 2 3   - - 6
	.byte	0x4f	! W 	0 1 2 3   - - 6
	.byte	0x70	! X		- - - -   4 5 6
	.byte	0x6d	! Y 	0 - 2 3   - 5 6
	.byte	0x7a	! Z 	- 1 - 3   4 5 6
	.byte	0x53	! [ 	0 1 - -   4 - 6
	.byte	0x25	! \ 	0 - 2 -   - 5 -
	.byte	0x5c	! ] 	- - 2 3   4 - 6
	.byte	0x10	! ^ 	- - - -   4 - -
	.byte	0x40	! _ 	- - - -   - - 6
	.byte	0x01	! ` 	0 - - -   - - -
	.byte	0x26	! a 	- 1 2 -   - 5 -
	.byte	0x67	! b 	0 1 2 -   - 5 6
	.byte	0x62	! c 	- 1 - -   - 5 6
	.byte	0x6e	! d 	- 1 2 3   - 5 6
	.byte	0x62	! e 	- 1 - -   - 5 6
	.byte	0x33	! f 	0 1 - -   4 5 -
	.byte	0x7d	! g 	0 - 2 3   4 5 6
	.byte	0x27	! h		0 1 2 -   - 5 -
	.byte	0x04	! i 	- - 2 -   - - -
	.byte	0x44	! j 	- - 2 -   - - 6
	.byte	0x27	! k 
	.byte	0x42	! l		- 1 - -   - - 6
	.byte	0x26	! m		- 1 2 -   - 5 -
	.byte	0x26	! n		- 1 2 -   - 5 -
	.byte	0x66	! o 	- 1 2 -   - 5 6
	.byte	0x3b	! p 	0 1 - 3   4 5 -
	.byte	0x3d	! q 	0 - 2 3   4 5 -
	.byte	0x22	! r		- 1 - -   - 5 -
	.byte	0x75	! s 	0 - 2 -   4 5 6
	.byte	0x63	! t		0 1 - -   - 5 6
	.byte	0x46	! u 	- 1 2 -   - - 6
	.byte	0x46	! v 	- 1 2 -   - - 6
	.byte	0x46	! w		- 1 2 -   - - 6
	.byte	0x06	! x		- 1 2 -   - - -
	.byte	0x2d	! y 	0 - 2 3   - 5 -
	.byte	0x60	! z 	- - - -   - 5 6
	.byte	0x62	! { 	- 1 - -   - 5 6
	.byte	0x02	! | 	- 1 - -   - - -
	.byte	0x64	! } 	- - 2 -   - 5 6
	.byte	0x10	! ~ 	- - - -   4 - -
	.byte	0x00	! ^? 	- 1 - 3   4 5 6

	fontsize = . - font


dpydiv13:
	ld c, -1
1:
	inc c
	sub 13
	jr nc, 1b
	ld a, c
	jp dpyval99

dpydiv16:
	srl4
	and 0xF
	jp dpydig

!	digit or character in a
!	location table in de
!	de incremented to next position,
!	hl & a destroyed

dpydig:

	push bc

	ld h, HI(font)	; ASSERT(LO(font) == 0)
	ld l, a
	ld c, [hl]		! what segments (0..6) are lit, what are dim

	ld a, [de]; inc de; rr c; call segment_onoff
	ld a, [de]; inc de; rr c; call segment_onoff
	ld a, [de]; inc de; rr c; call segment_onoff
	ld a, [de]; inc de; rr c; call segment_onoff

	ld a, [de]; inc de; rr c; call segment_onoff
	ld a, [de]; inc de; rr c; call segment_onoff
	ld a, [de]; inc de; rr c; call segment_onoff

	pop bc
	ret

tx_on:
	di
	ld a, 1
	ld [txon], a
	ld a, 0x04
	out [TMR + 1], a
	ld a, 0x00
	out [TMR + 1], a
	ld a, [output_0]
	and ~ O0_AUDIOC
and ~ O0_MTC
and ~ O0_CCIRC
	ld [output_0], a
	out [OUT0], a
	ei
#if 1
	load_abc(tx_sub)
	ld e, a
	load_ahl(tx_divisor)
	and a					! clear carry
	sbc hl, bc
	sub e
	save_ahl(tx_divisor)
#endif
	ld a, O1_TXOFF
	out [OUT1], a
	ld a, 0
	out [DA_TXPWR], a
	call txmode_synth
	call change_synth
	ld a, 0
	out [OUT1], a
	ld a, [txpwr]	!  0  1  2  3
	sla a			!  0  2  4  6
	sla a			!  0  4  8  C
	sla a			!  0  8 10 18
	sla a			!  0 10 20 30
	sla a			!  0 20 40 60
	sla a			!  0 40 80 C0
	or 0x3F			! 3F 7F BF FF
	
	out [DA_TXPWR], a
	ret

tx_off:
	ld a, O1_TXOFF
	out [OUT1], a
	ld a, 0
	out [DA_TXPWR], a
#if 1
	load_abc(tx_sub)
	ld e, a
	load_ahl(tx_divisor)
	add hl, bc
	add e
	save_ahl(tx_divisor)
#endif
	call synth_setup
	call change_synth

	ld a, 0
	ld [txon], a

	ret

	!
	!  Calculate binary value from unpacked string
	!  Return in CY AHL
	!
a2i:
	push ix
	push bc
	push de

	ld ix, digbuf
	ld a, [digidx]
	ld c, a

	ld a, 0				! AHL starts from 0x000000
	ld h, a
	ld l, a
	ld [digidx], a

	dec c
	jp z, a2i1
	dec c
	jp z, a2i2
	dec c
	jp z, a2i3
	dec c
	jp z, a2i4
	dec c
	jp z, a2i5
	dec c
	jp z, a2i6
	dec c
	jp z, a2i7

	jp a2i0				! more than 7, give 0

a2iOV:
	imm_ahl(0xFFFFFF)
	jp a2i0

a2i7:
	ld de, 0x4240
	ld b, [ix]		! x 000 000
	inc ix
	inc b
	jr 1f
2:
	add hl, de
	adc 0x0F			! 1 000 000 equals 0x 0F 4240
	jp c, a2iOV
1:
	djnz 2b
a2i6:
	ld de, 0x86a0		! 100 000 equals 0x186a0, thus the adc 1...
	ld b, [ix]		! 0 x00 000
	inc ix
	inc b
	jr 1f
2:
	add hl, de
	adc 1
	jp c, a2iOV
1:
	djnz 2b
a2i5:
	ld de, 10000
	ld b, [ix]		! 0x0 000
	inc ix
	inc b
	jr 1f
2:
	add hl, de
	adc 0
	jp c, a2iOV
1:
	djnz 2b
a2i4:
	ld de, 1000
	ld b, [ix]		! 00x 000
	inc ix
	inc b
	jr 1f
2:
	add hl, de
	adc 0
	jp c, a2iOV
1:
	djnz 2b
a2i3:
	ld de, 100
	ld b, [ix]		! 000 x00
	inc ix
	inc b
	jr 1f
2:
	add hl, de
	adc 0
	jp c, a2iOV
1:
	djnz 2b
a2i2:
	ld de, 10
	ld b, [ix]		! 000 0x0
	inc ix
	inc b
	jr 1f
2:
	add hl, de
	adc 0
	jp c, a2iOV
1:
	djnz 2b
a2i1:
	ld de, 1
	ld b, [ix]		! 000 00x
	inc ix
	inc b
	jr 1f
2:
	add hl, de
	adc 0
	jp c, a2iOV
1:
	djnz 2b
a2i0:
	pop de
	pop bc
	pop ix

	ret					! AHL contains frequency in binary


!	Multiply AHL by C, 1..
!	Result to CY AHL

mul248:

	push bc
	push de

	ld e, l
	ld d, h
	ld b, a
	and a			! clear carry
1:
	dec c
	jr z, 1f
	adc hl, de
	adc b
	jr nc, 1b
	imm_ahl(0xFFFFFF)
1:
	pop de
	pop bc

	ret

!----------------------------------------------------------------------
!
! Divide AHL by C, result left in HL, remainder in A.
!
! Origin: Trash-80 Assembly Language Subroutines, William Barden 1982.
!

div248:

	push bc

	ld b, 16	! do this 16 times.
2:
	add hl, hl	! rotate AHL one bit
	adc a		! position to the left.
	sub c
	jr c, 1f
	inc hl		! set lsb of hl, a bit into quotient.
	djnz 2b
	jr 3f
1:
	add c		! oops, should not have subtracted.
	djnz 2b
3:
	pop bc

	ret

	!
	!	freq / 25 * 2 = xxx
	!	AHL has freq in Hz
	!

freq2div:

	push bc
	push de

	ld bc, [bstep_cfg]
	call div248
	ld e, 0				! MSb
	inc b
	jr 2f
1:
	scf					! remainder * 2 + 1
	rl a				! 
	cp c				! compare with divider
	jr c, 3f
	sub c
3:
	ccf
	rl l				! and shift in
	rl h
	rl e
2:
	djnz 1b

	ld a, e
	cp 2				! 0 or 1 only possible
	jr c, 1f
	imm_ahl(0x1FFFF)
1:

	pop de
	pop bc

	ret

	!	Print tx_divisor as Hz, 7 digits, at de
	!
	!	xxx * 25 / 2 = freq
	!
div2dpy:

	push bc
	push de

	ld a, ' '
	ld [zeroblank], a

	load_ahl(tx_divisor)

	ld bc, [bstep_cfg]
	call mul248
	inc b
	jr 2f
1:
	rr a
	rr h
	rr l
	and a
2:
	djnz 1b

	ld de, 10000
	call divide		! result to c

	pop de			! get cursor

	push af

	ld a, c			! 10 MHz and up digits

	ld c, -1
1:
	inc c			! count GHz digit
	sub 100
	jr nc, 1b
	add 100
	call nextdig	! insert GHz

	ld c, -1
1:
	inc c			! count 100 MHz digit
	sub 10
	jr nc, 1b
	add 10
	call nextdig	! insert 100 MHz

	ld c, a
	call nextdig	! insert 10 MHz

	pop af

	push de			! display cursor

	ld de, 1000
	call divide

	pop de
	call nextdig
	push de

	ld de, 100
	call divide

	pop de
	call nextdig
	push de

	ld de, 10
	call divide

	pop de
	call nextdig

	ld a, 0
	ld [zeroblank], a

	ld c, l
	call nextdig

	pop bc

	ret

divide:
	ld c, -1
1:
	inc c
	and a
	sbc hl, de
	sbc 0
	jr nc, 1b
	add hl, de
	adc 0
	ret

nextdig:
	push hl
	push af
	ld a, c
	or a
	jr z, 1f
	call dpydig
	ld a, 0
	ld [zeroblank], a
	pop af
	pop hl
	ret
1:
	ld a, [zeroblank]
	call dpydig
	pop af
	pop hl
	ret

div2freq:

	push bc

	ld bc, [bstep_cfg]
	call mul248
	inc b
	jr 2f
1:
	rr a
	rr h
	rr l
	and a
2:
	djnz 1b

	pop bc

	ret

!======================================================================

	.asciz "TheEnd"

1:
	.data

#ifdef PROM
	.org 0xC000		! start of RAM
#else
	.org 1b			! Trailing RAM-based code
#endif

!======================================================================
!
!	Variables
!

output_1	BYTE

mbusrx_rp	WORD
mbusrx_wp	WORD

mbustx_rp	WORD
mbustx_wp	WORD

output_0	BYTE

txon		BYTE
volume		BYTE
rfc			BYTE
rssi		BYTE
srssi		BYTE
txpwr		BYTE
batt		BYTE

squelch_limit_close	BYTE
squelch_limit_open	BYTE

currmem		BYTE
currfreq	BUF(3)	! current frequency in kHz
tx_sub		BUF(3)	! repeater difference
rx_sub		BUF(3)	! for the 86512.5 kHz
tx_divisor	BUF(3)	! 17 bits

memories:	.rs (4 * 128)

!======================================================================

	.align 8			! zeroed variables

	ASSERT(LO(.) == 0)

_bss:

mbusrx_buf	BUF(256)	! page aligned
mbustx_buf	BUF(256)	! page aligned

segments	BUF(16)		! 32 bits 4 times, page aligned
indicators	BYTE		! trails segments !

mbustx_cnt	BYTE
mbusrx_cnt	BYTE

mbus_timer	BYTE		! timer for MBUS transmit

rssi_timer	BYTE
mt_timer	BYTE
key_timer	BYTE
key_speed	BYTE

light_ltimer	WORD

sir			BYTE
KEYSIR = 0
DPYSIR = 1
INSIR  = 7

bstep_cfg	BUF(2)	! 2 and 25

bstep_R		WORD	! 1024

lastkey		BYTE
key			BYTE
dark		BYTE

pttdn		BYTE
keydown		BYTE

digidx		BYTE
digbuf		BUF(16)

zeroblank	BYTE

shift_state	BYTE

squelch_open	BYTE
squelch_forced	BYTE

_end:

